/*
 * pipefs - 
 * (c) 2003, François Revol, revol@free.fr
 */

/*
 * TODO:
 * - wstat()
 * - O_MKFIFO & mkfifo & unlink()
 */

#define _BUILDING_fs 1

#include <sys/param.h>
#include <sys/stat.h>
#include <malloc.h>
#include <KernelExport.h>
//#include <NodeMonitor.h>
#include <stddef.h>
#include <signal.h>
#include "lock.h"
#include "pipefs.h"

//#define PRINT(a) dprintf a
#define PRINT(a) 
#define PFS "pipefs: "

/* needed to get /bin/df tell the mountpoint... */
#define ALLOW_DIR_OPEN

_EXPORT vnode_ops fs_entry =
{
	(op_read_vnode *)&pipefs_read_vnode,
	(op_write_vnode *)&pipefs_write_vnode,
	(op_remove_vnode *)&pipefs_remove_vnode,
	(op_secure_vnode *)&pipefs_secure_vnode,
	(op_walk *)&pipefs_walk,
	(op_access *)&pipefs_access,
	(op_create *)&pipefs_create,
	NULL, //(op_mkdir *)&pipefs_mkdir,
	NULL, //(op_symlink *)&pipefs_symlink,
	NULL, //	&pipefs_link,
	NULL, //(op_rename *)&pipefs_rename,
	(op_unlink *)&pipefs_unlink,
	NULL, //(op_rmdir *)&pipefs_rmdir,
	NULL, //(op_readlink *)&pipefs_readlink,
	(op_opendir *)&pipefs_opendir,
	(op_closedir *)&pipefs_closedir,
	(op_free_cookie *)&pipefs_free_dircookie,
	(op_rewinddir *)&pipefs_rewinddir,
	(op_readdir *)&pipefs_readdir,
	(op_open *)&pipefs_open,
	(op_close *)&pipefs_close,
	(op_free_cookie *)&pipefs_free_cookie,
	(op_read *)&pipefs_read,
	(op_write *)&pipefs_write,
	NULL, //	&pipefs_readv,
	NULL, //	&pipefs_writev,
	NULL, //	&pipefs_ioctl,
	(op_setflags *)&pipefs_setflags,
	(op_rstat *)&pipefs_rstat,
	(op_wstat *)&pipefs_wstat,
	NULL, //	&pipefs_fsync,
	NULL, //	&pipefs_initialize,
	(op_mount *)&pipefs_mount,
	(op_unmount *)&pipefs_unmount,
	NULL, //	&pipefs_sync,
	(op_rfsstat *)&pipefs_rfsstat,
	(op_wfsstat *)&pipefs_wfsstat,
	(op_select *)&pipefs_select,
	(op_deselect *)&pipefs_deselect,
	NULL, //	&pipefs_open_indexdir,
	NULL, //	&pipefs_close_indexdir,
	NULL, //	&pipefs_free_indexdircookie,
	NULL, //	&pipefs_rewind_indexdir,
	NULL, //	&pipefs_read_indexdir,
	NULL, //	&pipefs_create_index,
	NULL, //	&pipefs_remove_index,
	NULL, //	&pipefs_rename_index,
	NULL, //	&pipefs_stat_index,
	NULL, //	&pipefs_open_attrdir,
	NULL, //	&pipefs_close_attrdir,
	NULL, //	&pipefs_free_attrdircookie,
	NULL, //	&pipefs_rewind_attrdir,
	NULL, //	&pipefs_read_attrdir,
	NULL, //	&pipefs_write_attr,
	NULL, //	&pipefs_read_attr,
	NULL, //	&pipefs_remove_attr,
	NULL, //	&pipefs_rename_attr,
	NULL, //	&pipefs_stat_attr,
	NULL, //	&pipefs_open_query,
	NULL, //	&pipefs_close_query,
	NULL, //	&pipefs_free_querycookie,
	NULL, //	&pipefs_read_query	
};

_EXPORT int32 api_version=B_CUR_FS_API_VERSION;

/* namespace MUST be locked */
vnode_id new_vnid(fs_nspace *ns)
{
	vnode_id vnid = ns->nextid++;
	if (vnid < 0) {
		panic(PFS "WE ARE BORK: nsid overflow!");
		vnid = ns->nextid++;
	}
	return vnid;
	/* XXX: handle overflow someday */
}

static void fill_default_stat(struct stat *st, nspace_id nsid, vnode_id vnid, mode_t mode)
{
	st->st_dev = nsid;
	st->st_ino = vnid;
	st->st_mode = mode;
	st->st_nlink = 1;
	st->st_uid = 0;
	st->st_gid = 0;
	st->st_size = 0LL;
	st->st_blksize = PIPEFS_BUFF_SIZE;
	st->st_atime = time(NULL);
	st->st_mtime = time(NULL);
	st->st_ctime = time(NULL);
	st->st_crtime = time(NULL);
}

int pipefs_mount(nspace_id nsid, const char *devname, ulong flags,
		struct mount_fs_params *parms, size_t len, fs_nspace **data, vnode_id *vnid)
{
	fs_nspace *ns;
	int err;
	PRINT((PFS "mount(%d, %s, 0x%08lx, %p, %ld, , )\n", nsid, devname, flags, parms, len));
	ns = malloc(sizeof(fs_nspace));
	if (!ns)
		return B_NO_MEMORY;
	memset(ns, 0, sizeof(fs_nspace));
	ns->nsid = nsid;
	ns->rootid = 0x1LL;
	ns->nextid = 0x2LL;
	err = new_lock(&(ns->l), "pipefs main lock");
	if (err < 0)
		return err;
	ns->first = NULL;
	
	/* create root dir */
	err = B_NO_MEMORY;
	ns->first = malloc(sizeof(fs_node));
	if (ns->first) {
		ns->root = ns->first;
		memset(ns->first, 0, sizeof(fs_node));
		ns->first->lent.next = NULL;
		strcpy(ns->first->name, ".");
		ns->first->vnid = ns->rootid;
		fill_default_stat(&ns->first->st, ns->nsid, ns->rootid, 0777 | S_IFDIR);
		ns->first->is_perm = true;
		err = new_lock(&(ns->first->l), "pipefs root dir");
		if (err >= 0) {
			ns->first->readers = NULL;
			ns->first->writers = NULL;
			ns->first->waiting_readers = NULL;
			ns->first->waiting_writers = NULL;
			rb_init(&ns->first->buffer, 0);
			*data = ns;
			*vnid = ns->rootid;
			err = new_vnode(nsid, *vnid, ns->first);
			if (err == B_OK) {
				PRINT((PFS "mount() OK, nspace@ %p, root@ %p\n", ns, ns->first));
				return B_OK;
			}
			free_lock(&ns->first->l);
		}
		free(ns->first);
	}
	free_lock(&ns->l);
	return err;
}

int pipefs_unmount(fs_nspace *ns)
{
	status_t err;
	struct fs_node *node;
	PRINT((PFS "unmount(%d)\n", ns->nsid));
	err = LOCK(&ns->l);
	if (err)
		return err;
	/* anything in still in use ? */
	for (node = ns->first; node; node = node->lent.next) {
		if (node->readers || node->writers || node->waiting_readers || node->waiting_writers) {
			UNLOCK(&ns->l);
			return B_NAME_IN_USE;
		}
	}
	for (node = ns->first; node; node = ns->first) {
		ns->first = node->lent.next; /* better cache that before we free node */
		free_lock(&node->l);
		free(node);
	}
	free_lock(&ns->l);
	free(ns);
	return B_OK;
}

static int compare_fs_node_by_vnid(fs_node *node, vnode_id *id)
{
	return !(node->vnid == *id);
}

int pipefs_remove_vnode(fs_nspace *ns, fs_node *node, char r)
{
	status_t err = B_OK;
	PRINT((PFS "remove_vnode(%d, %Ld, %s)\n", ns->nsid, node->vnid, r?"r":"!r"));
	if (!r)
		err = LOCK(&ns->l);
	if (err)
		return err;
	if (node->vnid == ns->rootid) {
		PRINT((PFS "asked to remove the root node!!\n"));
	}
	free_lock(&node->l);
	err = slist_remove(&ns->first, node);
	free(node);
	if (!r)
		UNLOCK(&ns->l);
	return err;
}

int pipefs_read_vnode(fs_nspace *ns, vnode_id vnid, char r, fs_node **node)
{
	fs_node *n;
	status_t err = B_OK;
	PRINT((PFS "read_vnode(%d, %Ld, %s)\n", ns->nsid, vnid, r?"r":"!r"));
	if (!r)
		err = LOCK(&ns->l);
	if (err)
		return err;
	n = (fs_node *)slist_find(ns->first,
				  (slist_compare_func)compare_fs_node_by_vnid,
				  (void *)&vnid);
	if (n)
		*node = n;
	else
		err = ENOENT;
	if (!r)
		UNLOCK(&ns->l);
	return err;
}

int pipefs_write_vnode(fs_nspace *ns, fs_node *node, char r)
{
	PRINT((PFS "write_vnode(%d, %Ld, %s)\n", ns->nsid, node->vnid, r?"r":"!r"));
	return B_OK;
}

int pipefs_secure_vnode(fs_nspace *ns, fs_node *node)
{
	PRINT((PFS "secure_vnode(%d, %Ld)\n", ns->nsid, node->vnid));
	return B_OK;
}

static int compare_fs_node_by_name(fs_node *node, char *name)
{
	//return memcmp(node->name, name, PIPEFS_NAME_LEN);
	PRINT((PFS"'%s' <> '%s'\n", node->name, name));
	return strncmp(node->name, name, PIPEFS_NAME_LEN);
}

int pipefs_walk(fs_nspace *ns, fs_node *base, const char *file, char **newpath, vnode_id *vnid)
{
	fs_node *n, *dummy;
	status_t err = B_OK;
	PRINT((PFS "walk(%d, %Ld, %s)\n", ns->nsid, (int64)(base?base->vnid:-1), file));
	err = LOCK(&ns->l);
	if (err)
		return err;
	if (!file || (strcmp(file, "..") == 0)) { /* parent */
		err = EINVAL;
	} else if (strcmp(file, ".") == 0) { /* root dir */
		*vnid = ns->rootid;
	} else if (base && (base->vnid == ns->rootid)) { /* child of root dir */
		n = (fs_node *)slist_find(ns->first,
					  (slist_compare_func)compare_fs_node_by_name,
					  (void *)file);
		if (n)
			*vnid = n->vnid;
		else
			err = ENOENT;
	} else
		err = ENOENT;
	if (err == B_OK)
		err = get_vnode(ns->nsid, *vnid, (void **)&dummy); /* inc ref count */
	UNLOCK(&ns->l);
	PRINT((PFS "walk() -> error 0x%08lx\n", err));
	return err;
}

int pipefs_opendir(fs_nspace *ns, fs_node *node, fs_dir_cookie **cookie)
{
	status_t err = B_OK;
	fs_dir_cookie *c;
	PRINT((PFS "opendir(%d, %Ld)\n", ns->nsid, node->vnid));
	if (!node)
		return EINVAL;
	if (!S_ISDIR(node->st.st_mode))
		return B_NOT_A_DIRECTORY;
	err = LOCK(&node->l);
	if (err)
		return err;
	c = malloc(sizeof(fs_dir_cookie));
	if (c) {
		memset(c, 0, sizeof(fs_dir_cookie));
		c->omode = O_RDONLY;
		c->wflags = FS_WAIT_NONE;
		c->node = node;
		c->dir_current = 0;
		err = slist_insert_head(&node->readers, &c->lent);
		if (err == B_OK) {
			*cookie = c;
			UNLOCK(&node->l);
			return B_OK;
		}
		free(c);
	} else
		err = B_NO_MEMORY;
	UNLOCK(&node->l);
	return err;
}

int pipefs_closedir(fs_nspace *ns, fs_node *node, fs_dir_cookie *cookie)
{
	status_t err = B_OK;
	PRINT((PFS "closedir(%d, %Ld)\n", ns->nsid, node->vnid));
	err = LOCK(&node->l);
	if (err)
		return err;
	err = slist_remove(&node->readers, &cookie->lent);
	UNLOCK(&node->l);
	return err;
}

int pipefs_rewinddir(fs_nspace *ns, fs_node *node, fs_dir_cookie *cookie)
{
	PRINT((PFS "rewinddir(%d, %Ld)\n", ns->nsid, node->vnid));
	cookie->dir_current = 0;
	return B_OK;
}

int pipefs_readdir(fs_nspace *ns, fs_node *node, fs_dir_cookie *cookie, long *num, struct dirent *buf, size_t bufsize)
{
	status_t err = B_OK;
	fs_node *n = NULL;
	int index;
	PRINT((PFS "readdir(%d, %Ld) @ %d\n", ns->nsid, node->vnid, cookie->dir_current));
	if (!node || !cookie || !num || !*num || !buf || (bufsize < (sizeof(dirent_t)+PIPEFS_NAME_LEN)))
		return EINVAL;
	if (cookie->dir_current == 0) { /* .. */
		PRINT((PFS "readdir() giving ..\n"));
		/* the VFS will correct that anyway */
		buf->d_dev = ns->nsid;
		buf->d_pdev = ns->nsid;
		buf->d_ino = ns->rootid;
		buf->d_pino = ns->rootid;
		strcpy(buf->d_name, "..");
		buf->d_reclen = 2*(sizeof(dev_t)+sizeof(ino_t))+sizeof(unsigned short)+strlen(buf->d_name)+1;
		cookie->dir_current++;
		*num = 1;
		return B_OK;
	}
	err = LOCK(&ns->l);
	index = cookie->dir_current-1;
	for (n = ns->first; n && index; n = n->lent.next, index--);
	if (n) {
		PRINT((PFS "readdir() giving ino %Ld, %s\n", n->vnid, n->name));
		buf->d_dev = ns->nsid;
		buf->d_pdev = ns->nsid;
		buf->d_ino = n->vnid;
		buf->d_pino = ns->rootid;
		strcpy(buf->d_name, n->name);
		buf->d_reclen = 2*(sizeof(dev_t)+sizeof(ino_t))+sizeof(unsigned short)+strlen(buf->d_name)+1;
		cookie->dir_current++;
		*num = 1;
	} else {
		*num = 0;
	}
	UNLOCK(&ns->l);
	return B_OK;
}

int pipefs_free_dircookie(fs_nspace *ns, fs_node *node, fs_dir_cookie *cookie)
{
	PRINT((PFS"freedircookie(%d, %Ld)\n", ns->nsid, node->vnid));
	free(cookie);
	return B_OK;
}

int pipefs_rstat(fs_nspace *ns, fs_node *node, struct stat *st)
{
	status_t err = B_OK;
	if (!node || !st)
		return EINVAL;
	err = LOCK(&node->l);
	if (err)
		return err;
	memcpy(st, &node->st, sizeof(struct stat));
	st->st_dev = ns->nsid;
	st->st_ino = node->vnid;
	st->st_size = rb_can_read(&node->buffer);
	UNLOCK(&node->l);
	return err;
}

int pipefs_rfsstat(fs_nspace *ns, struct fs_info *info)
{
	info->block_size=0;//PIPEFS_BUFF_SIZE;
	info->io_size=0;//PIPEFS_BUFF_SIZE;
	info->total_blocks=0;
	info->free_blocks=0;
	info->total_nodes=0;
	info->free_nodes=0;
	info->dev=ns->nsid;
	info->root=ns->rootid;
	info->flags=0;
	strcpy (info->device_name, "");
	strcpy (info->volume_name, "");
	strcpy (info->fsh_name, PIPEFS_NAME);
	return B_OK;
}

int pipefs_open(fs_nspace *ns, fs_node *node, int omode, fs_file_cookie **cookie)
{
	status_t err = B_OK;
	fs_node *dummy;
	fs_file_cookie *fc;
	PRINT((PFS"open(%d, %Ld, 0x%x)\n", ns->nsid, node->vnid, omode));
	if (!node || !cookie)
		return EINVAL;
	/* can only be read OR write */
	if ((omode & O_RWMASK) == O_RDWR)
		return EINVAL;
//	err = LOCK(&ns->l);
//	if (err)
//		return err;
	err = LOCK(&node->l);
	if (err)
		goto err_n_l;
	err = EEXIST;
	if ((omode & O_EXCL) && ((omode & O_RWMASK) == O_RDONLY) && node->readers)
		goto err_malloc;
	if ((omode & O_EXCL) && ((omode & O_RWMASK) == O_WRONLY) && node->writers)
		goto err_malloc;
	if (((omode & O_RWMASK) == O_RDONLY) && node->readers && (node->readers->omode & O_EXCL))
		goto err_malloc;
	if (((omode & O_RWMASK) == O_WRONLY) && node->writers && (node->writers->omode & O_EXCL))
		goto err_malloc;
#ifndef ALLOW_DIR_OPEN
	err = EINVAL;//EISDIR;
	if (S_ISDIR(node->st.st_mode))
		goto err_malloc;
#endif
	err = B_NO_MEMORY;
	fc = malloc(sizeof(fs_file_cookie));
	if (!fc)
		goto err_malloc;
	memset(fc, 0, sizeof(fs_file_cookie));
	slist_init(&fc->lent);
	slist_init(&fc->wlent);
	fc->node = node;
	fc->omode = omode;
	fc->wflags = FS_WAIT_NONE;
	fc->sem = create_sem(0, "pipe sem");
	if (fc->sem < B_OK)
		goto err_sem;
	set_sem_owner(fc->sem, B_SYSTEM_TEAM);
	err = slist_insert_head(((omode & O_RWMASK) == O_RDONLY)?(&node->readers):(&node->writers), fc);
	if (err)
		goto err_linsert;
	if ((omode & O_RWMASK) == O_WRONLY)
		node->got_wr = true;
/*	err = get_vnode(ns->nsid, node->vnid, &dummy);
	if (err)
		goto err_getvn;*/
	//*vnid = node->vnid;
	*cookie = fc;
	err = B_OK;
	goto all_ok;
//err_:
//	put_vnode(ns->nsid, node->nsid);
err_getvn:
	slist_remove(((omode & O_RWMASK) == O_RDONLY)?(&node->readers):(&node->writers), fc);
err_linsert:
	delete_sem(fc->sem);
err_sem:
	free(fc);
err_malloc:
all_ok:
	UNLOCK(&node->l);
err_n_l:
//	UNLOCK(&ns->l);
	return err;
}

int pipefs_close(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie)
{
	status_t err;
	bool r; /* read/write */
	int i;
	PRINT((PFS"close(%d, %Ld)\n", ns->nsid, node->vnid));
	if (!ns || !node || !cookie)
		return EINVAL;
	err = LOCK(&node->l);
	if (err)
		return err;
	r = ((cookie->omode & O_RWMASK) == O_RDONLY);
	slist_remove(r?(&node->waiting_readers):(&node->waiting_writers), &cookie->wlent);
	if (cookie->sem > B_OK)
		delete_sem(cookie->sem);
	for (i = 0; i < SELECT_MAX; i++) {
		if (cookie->selsync[i])
			notify_select_event(cookie->selsync[i], cookie->selref[i]);
	}
	cookie->wflags = FS_WAIT_NONE;
	
all_ok:
err_n_l:
	UNLOCK(&node->l);
	return err;
}

int pipefs_free_cookie(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie)
{
	status_t err = B_OK;
	bool r; /* read/write */
	PRINT((PFS"freecookie(%d, %Ld)\n", ns->nsid, node->vnid));
	r = ((cookie->omode & O_RWMASK) == O_RDONLY);
	err = LOCK(&node->l);
	if (err)
		return err;
	err = slist_remove(r?(&node->readers):(&node->writers), cookie);
//	if (err)
//		goto err_n_l;
	if ((node->readers == NULL) && (node->writers == NULL)) {
		rb_clear(&node->buffer);
		node->got_wr = false;
		/* destroy the node */
		if (!node->is_perm) {
			err = remove_vnode(ns->nsid, node->vnid);
			ns->root->st.st_mtime = time(NULL);
			notify_listener(B_ENTRY_REMOVED, ns->nsid, ns->rootid, 0LL, node->vnid, NULL);
			notify_listener(B_STAT_CHANGED, ns->nsid, 0LL, 0LL, ns->rootid, NULL);
		} else
			UNLOCK(&node->l);
	} else {
		if (r && (node->readers == NULL))
			pipefs_event(ns, node, FS_NOTIFY_ALL|FS_NOTIFY_WRITER);
		if (!r && (node->writers == NULL))
			pipefs_event(ns, node, FS_NOTIFY_ALL|FS_NOTIFY_READER);
		UNLOCK(&node->l);
	}
	free(cookie);
//	err = B_OK;
err_n_l:
	return err;
}

int pipefs_read(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, off_t pos, void *buf, size_t *len)
{
	status_t err = B_OK;
	size_t toread = *len;
	size_t done = 0;
	size_t pass = 0;
	PRINT((PFS"read(%d, %Ld, %Ld, %ld)\n", ns->nsid, node->vnid, pos, *len));
	err = EINVAL;
	if ((cookie->omode & O_RWMASK) == O_WRONLY)
		goto err;
	err = EINVAL;//EISDIR;
	if (S_ISDIR(node->st.st_mode))
		goto err;
	while (done < toread) {
		err = LOCK(&node->l);
		if (err)
			goto err;
		if (rb_can_read(&node->buffer)) {
			pass = rb_read(&node->buffer, buf, toread);
			if (pass) {
				done += pass;
				buf += pass;
				toread -= pass;
			} else
				goto err_n_l;
		}
		if (done)
			pipefs_event(ns, node, FS_NOTIFY_WRITER);
		if (node->got_wr && !node->writers) /* EOF */
			goto err_n_l;
		if (done) /* got something */
			goto err_n_l;
		err = EAGAIN;
		if (cookie->omode & O_NONBLOCK)
			goto err_n_l;
		if (!cookie->wflags)
			slist_insert_head(&node->waiting_readers, &cookie->wlent);
		cookie->wflags |= FS_WAIT_SEM;
		
		UNLOCK(&node->l);
		err = acquire_sem_etc(cookie->sem, 1, B_CAN_INTERRUPT, 0LL);
		if (err)
			goto err;
	}
err_n_l:
	UNLOCK(&node->l);
err:
	*len = done;
	PRINT((PFS"read: len %ld, error 0x%08lx\n", done, err));
	return err;
}

int pipefs_write(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, off_t pos, const void *buf, size_t *len)
{
	status_t err = B_OK;
	size_t towrite = *len;
	size_t done = 0;
	size_t pass = 0;
	PRINT((PFS"write(%d, %Ld, %Ld, %ld)\n", ns->nsid, node->vnid, pos, *len));
	err = EINVAL;
	if ((cookie->omode & O_RWMASK) == O_RDONLY)
		goto err;
	err = EINVAL;//EISDIR;
	if (S_ISDIR(node->st.st_mode))
		goto err;
	while (done < towrite) {
		err = LOCK(&node->l);
		if (err)
			goto err;
		if (!done && !node->readers) /* broken pipe */
			goto err_epipe;
		if (rb_can_write(&node->buffer)) {
			pass = rb_write(&node->buffer, buf, towrite);
			if (pass) {
				done += pass;
				buf += pass;
				towrite -= pass;
			} else
				goto err_n_l;
		}
		if (done)
			pipefs_event(ns, node, FS_NOTIFY_READER);
		if (done)/* got something */
			goto err_n_l;
		err = EAGAIN;
		if (cookie->omode & O_NONBLOCK)
			goto err_n_l;
		if (!cookie->wflags)
			slist_insert_head(&node->waiting_writers, &cookie->wlent);
		cookie->wflags |= FS_WAIT_SEM;
		
		UNLOCK(&node->l);
		err = acquire_sem_etc(cookie->sem, 1, B_CAN_INTERRUPT, 0LL);
		if (err)
			goto err;
	}
err_epipe:
	err = EPIPE;
	kill(find_thread(NULL), SIGPIPE);
err_n_l:
	UNLOCK(&node->l);
err:
	*len = done;
	PRINT((PFS"write: len %ld, error 0x%08lx\n", done, err));
	return err;
}

int pipefs_wstat(fs_nspace *ns, fs_node *node, struct stat *st, long mask)
{
	PRINT((PFS"wstat(%d, %Ld, , 0x%08lx)\n", ns->nsid, node->vnid, mask));
	return ENOSYS;
}

int pipefs_wfsstat(fs_nspace *ns, struct fs_info *info, long mask)
{
	PRINT((PFS"wfsstat(%d, , 0x%08lx)\n", ns->nsid, mask));
	return ENOSYS;
}

int pipefs_create(fs_nspace *ns, fs_node *dir, const char *name, int omode, int perms, vnode_id *vnid, fs_file_cookie **cookie)
{
	status_t err = B_OK;
        fs_node *n = NULL;
	fs_node *dummy;
	fs_file_cookie *fc;
	bool created = false;
	PRINT((PFS"create(%d, %Ld, %s, 0x%x, 0%4o, , )\n", ns->nsid, dir->vnid, name, omode, perms));
	if (!dir || !name || !vnid || !cookie)
		return EINVAL;
	if (strlen(name) >= PIPEFS_NAME_LEN)
		return ENAMETOOLONG;
	/* can only be read OR write */
	if ((omode & O_RWMASK) == O_RDWR)
		return EINVAL;
	err = LOCK(&ns->l);
	if (err)
		return err;
	n = (fs_node *)slist_find(ns->first,
				(slist_compare_func)compare_fs_node_by_name,
				(void *)name);
	if (n) {
		if (omode & O_EXCL) {
			UNLOCK(&ns->l);
			return EEXIST;
		}
	} else { /* no node yet, create it */
		created = true;
		err = B_NO_MEMORY;
		n = malloc(sizeof(fs_node)+PIPEFS_BUFF_SIZE+0x10); /* extra; paranoid */
		if (!n)
			goto err_mallocn;
		memset(n, 0, sizeof(fs_node));
		//memset(((char *)n)+sizeof(fs_node)+PIPEFS_BUFF_SIZE, 0x5a, 0x10);
		strcpy(n->name, name);
		err = new_lock(&n->l, "a pipe's lock");
		if (err)
			goto err_newlk;
		n->vnid = new_vnid(ns);
		fill_default_stat(&n->st, ns->nsid, n->vnid, 0666 | S_IFIFO);
		rb_init(&n->buffer, PIPEFS_BUFF_SIZE);
		err = new_vnode(ns->nsid, n->vnid, n);
		if (err)
			goto err_newvnid;
		err = slist_insert_head(&ns->first, n);
		if (err)
			goto err_linsert;
		
		ns->root->st.st_mtime = time(NULL);
		notify_listener(B_ENTRY_CREATED, ns->nsid, ns->rootid, 0LL, n->vnid, name);
		notify_listener(B_STAT_CHANGED, ns->nsid, 0LL, 0LL, ns->rootid, NULL);
	}
	err = LOCK(&n->l);
	if (err)
		goto err_n_l;
	err = EEXIST;
	if (((omode & O_RWMASK) == O_RDONLY) && n->readers && (n->readers->omode & O_EXCL))
		goto err_mallocfc;
	if (((omode & O_RWMASK) == O_WRONLY) && n->writers && (n->writers->omode & O_EXCL))
		goto err_mallocfc;
	err = EISDIR;
	if (S_ISDIR(n->st.st_mode))
		goto err_mallocfc;
	err = B_NO_MEMORY;
	fc = malloc(sizeof(fs_file_cookie));
	if (!fc)
		goto err_mallocfc;
	memset(fc, 0, sizeof(fs_file_cookie));
	slist_init(&fc->lent);
	slist_init(&fc->wlent);
	fc->node = n;
	fc->omode= omode;
	fc->wflags = FS_WAIT_NONE;
	fc->sem = create_sem(0, "pipe sem");
	if (fc->sem < B_OK)
		goto err_sem;
	set_sem_owner(fc->sem, B_SYSTEM_TEAM);
	err = slist_insert_head(((omode & O_RWMASK) == O_RDONLY)?(&n->readers):(&n->writers), fc);
	if (err)
		goto err_linsertfc;
	if ((omode & O_RWMASK) == O_WRONLY)
		n->got_wr = true;
	if (!created)
		err = get_vnode(ns->nsid, n->vnid, &dummy);
	if (err)
		goto err_getvn;
	
	*vnid = n->vnid;
	*cookie = fc;
	UNLOCK(&n->l);
	UNLOCK(&ns->l);
	return B_OK;


err_getvn:
	slist_remove(((omode & O_RWMASK) == O_RDONLY)?(&n->readers):(&n->writers), fc);
err_linsertfc:
	delete_sem(fc->sem);
err_sem:
	free(fc);
err_mallocfc:
	UNLOCK(&n->l);
err_n_l:
	if (!created)
		return err;
	slist_remove(&ns->first, n);
err_linsert:
	put_vnode(ns->nsid, n->vnid);
err_newvnid:
	free_lock(&n->l);
err_newlk:
	free(n);
err_mallocn:
	UNLOCK(&ns->l);
	return err;
}

int pipefs_unlink(fs_nspace *ns, fs_node *dir, const char *name)
{
	PRINT((PFS"unlink(%d, %Ld, %s)\n", ns->nsid, dir->vnid, name));
	return ENOSYS;
}

int pipefs_access(fs_nspace *ns, fs_node *node, int mode)
{
	PRINT((PFS"access(%d, %Ld, 0x%x)\n", ns->nsid, node->vnid, mode));
	return B_OK;
}

int pipefs_setflags(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, int flags)
{
	status_t err = EINVAL;
	bool r; /* read/write */
	PRINT((PFS"setflags(%d, %Ld, 0x%x)\n", ns->nsid, node->vnid, flags));
	err = EINVAL;//EISDIR;
	if (S_ISDIR(node->st.st_mode))
		goto err;
	if ((flags & O_RWMASK) == O_RDWR)
		goto err;
	err = LOCK(&node->l);
	if (err)
		goto err;
	r = ((cookie->omode & O_RWMASK) == O_RDONLY);
	err = EINVAL;
	if (r && ((flags & O_RWMASK) == O_WRONLY))
		goto err_n_l;
	if (!r && ((flags & O_RWMASK) == O_RDONLY))
		goto err_n_l;
	flags &= (O_NONBLOCK);
	cookie->omode = (cookie->omode & ~(O_NONBLOCK)) | flags;
	err = B_OK;
err_n_l:
	UNLOCK(&node->l);
err:
	return err;
}

int pipefs_select(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, uint8 event, uint32 ref, selectsync *sync)
{
	status_t err = EINVAL;
	bool r; /* read/write */
	PRINT((PFS"select(%d, %Ld, %s, 0x%08lx, %ld)\n", ns->nsid, node->vnid, (event == SELECT_WRITE)?"w":((event == SELECT_READ)?"r":"e"), sync, ref));
	if (((cookie->omode & O_RWMASK) == O_RDONLY) && (event == SELECT_WRITE))
		goto err;
	if (((cookie->omode & O_RWMASK) == O_WRONLY) && (event == SELECT_READ))
		goto err;
	if (event > SELECT_EXCEPTION)
		goto err;
	err = LOCK(&node->l);
	if (err)
		goto err;
	err = EINPROGRESS;
	if (cookie->selsync[event])
		goto err_n_l;
	r = ((cookie->omode & O_RWMASK) == O_RDONLY);
	if (r && node->got_wr && !node->writers)
		notify_select_event(sync, ref);
	else if ((r && (event == SELECT_READ) && rb_can_read(&node->buffer)) || 
			(!r && (event == SELECT_WRITE) && rb_can_write(&node->buffer)))
		notify_select_event(sync, ref);
	else {
		cookie->selsync[event] = sync;
		cookie->selref[event] = ref;
		if (!cookie->wflags)
			slist_insert_head(r?(&node->waiting_readers):(&node->waiting_writers), &cookie->wlent);
		cookie->wflags |= FS_WAIT_SEL;
	}
		
	err = B_OK;
	PRINT((PFS"select ok\n"));
err_n_l:
	UNLOCK(&node->l);
err:
	return err;
}

int pipefs_deselect(fs_nspace *ns, fs_node *node, fs_file_cookie *cookie, uint8 event, selectsync *sync)
{
	status_t err = EINVAL;
	PRINT((PFS"deselect(%d, %Ld, %s, 0x%08lx)\n", ns->nsid, node->vnid, (event == SELECT_WRITE)?"w":((event == SELECT_READ)?"r":"e"), sync));
	if (((cookie->omode & O_RWMASK) == O_RDONLY) && (event == SELECT_WRITE))
		goto err;
	if (((cookie->omode & O_RWMASK) == O_WRONLY) && (event == SELECT_READ))
		goto err;
	if (event > SELECT_EXCEPTION)
		goto err;
	err = LOCK(&node->l);
	if (err)
		goto err;
	err = EINVAL;
	if (!cookie->selsync[event] || (cookie->selsync[event] != sync))
		goto err_n_l;
	cookie->selsync[event] = NULL;
	err = B_OK;
	PRINT((PFS"deselect ok\n"));
err_n_l:
	UNLOCK(&node->l);
err:
	return err;
}


/* make sure whatever the state of the node is, each reader/writer 
 * that should be is notified.
 * - there is room for write -> notify the first writer
 * - there is available data -> notify the first reader
 * - there is no reader -> notify all writers
 * - there is no writer -> notify all readers
 * node must be locked
 */
int pipefs_event(fs_nspace *ns, fs_node *node, int flags)
{
	fs_file_cookie *c;
	slist_entry *slent;
	PRINT((PFS"pipefs_event(%d, %Ld, %s%s%s)\n", ns->nsid, node->vnid, 
				(flags & FS_NOTIFY_ALL)?"all ":"", 
				(flags & FS_NOTIFY_READER)?"reader ":"", 
				(flags & FS_NOTIFY_WRITER)?"writer ":""));
	while ((flags & FS_NOTIFY_READER) && (slent = slist_dequeue_tail(&node->waiting_readers))) {
		c = ((char *)slent) - offsetof(fs_file_cookie, wlent);
		//PRINT((PFS"pipefs_event(), rlent=%08lx, cookie=%08lx\n", slent, c));
		if (c->wflags & FS_WAIT_SEM) {
			//PRINT((PFS"pipefs_event(r), releasing sem %d\n", c->sem));
			release_sem(c->sem);
		}
		if (c->wflags & FS_WAIT_SEL) {
			if (c->selsync[SELECT_READ])
				notify_select_event(c->selsync[SELECT_READ], c->selref[SELECT_READ]);
			if ((flags & FS_NOTIFY_ALL) && c->selsync[SELECT_EXCEPTION])
				notify_select_event(c->selsync[SELECT_EXCEPTION], c->selref[SELECT_EXCEPTION]);
		}
		c->wflags = FS_WAIT_NONE;
		if (!(flags & FS_NOTIFY_ALL))
			break;
	}
	while ((flags & FS_NOTIFY_WRITER) && (slent = slist_dequeue_tail(&node->waiting_writers))) {
		c = ((char *)slent) - offsetof(fs_file_cookie, wlent);
		//PRINT((PFS"pipefs_event(), rlent=%08lx, cookie=%08lx\n", slent, c));
		if (c->wflags & FS_WAIT_SEM) {
			//PRINT((PFS"pipefs_event(w), releasing sem %d\n", c->sem));
			release_sem(c->sem);
		}
		if (c->wflags & FS_WAIT_SEL) {
			if (c->selsync[SELECT_WRITE])
				notify_select_event(c->selsync[SELECT_WRITE], c->selref[SELECT_WRITE]);
			if ((flags & FS_NOTIFY_ALL) && c->selsync[SELECT_EXCEPTION])
				notify_select_event(c->selsync[SELECT_EXCEPTION], c->selref[SELECT_EXCEPTION]);
		}
		c->wflags = FS_WAIT_NONE;
		if (!(flags & FS_NOTIFY_ALL))
			break;
	}
	return B_OK;
}
